package ro.uvt.cs.ads;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Arrays;
import java.util.Scanner;

/**
 * The {@code kdTree} class represents a k-d tree.
 * @author Mircea Marin
 *
 */
public class kdTree {
	private final int n; // number of dimensions
	private Node root;
	
	public kdTree(int n) {
		this.n=n;
	}

		
	/**
	 * Returns the nearest neighbor to the test point {@code test} in the k-d tree {@code kd}
	 *  
	 * @param kd the k-d tree
	 * @param test the test point 
	 * @return the nearest neighbor
	 */
	public static double[] NN1(kdTree kd,double[] test) {
		return Node.NN1(kd.root,test);
	}
	
	/**
	 * Returns the k nearest neighbors to the test point {@code test} in the k-d tree {@code kd}
	 * 
	 * @param kd the k-d tree
	 * @param k the number of nearest neighbors to be found
	 * @param test the test point
	 * @return a bounded priority queue with the k nearest neighbors from in the k-d tree {@code kd}
	 */
	public static BPQ<double[]> NNk(kdTree kd, int k, double[] test) {
	    return Node.NNn(kd.root, test, k);
	}

	public void makeTree(double[][] points,int s,int e,int axis) {
		if(e<=s) return;
		if (e==s+1) {
			root = new Node(points[s]);
		} else {
			Arrays.sort(points,s,e,
					(x,y) -> (x[axis]<y[axis])?-1:(x[axis]==y[axis]?0:1));
			int m = (e+s)/2;
			while ((m+1<e)&&points[m][axis]==points[m+1][axis]) m++;
			root = new Node(points[s+m]);
			root.makeTree('l',points,s,m,(axis+1)%n);
			root.makeTree('r',points,m+1,e,(axis+1)%n);
		}
	}
	
	public String toString() { 
		StringBuffer sb = new StringBuffer(); 
		if(root!=null) root.toString(sb, "");
		return sb.toString();
	}
	
	public static void showNeighbor(double[] x) {
		System.out.print("(");
		for (int i=1;i<x.length;i++)
			System.out.print(String.format("%4.2f,",x[i-1]));
		System.out.println(String.format("%4.2f)",x[x.length-1]));
	}
	
	public static void main(String[] args) throws FileNotFoundException {
		File text = new File("/Users/mincea/eclipse-workspace/ADS/bin/ro/uvt/cs/ads/points.txt");
		Scanner myInput = new Scanner(text);
		
		int m = myInput.nextInt(); // number of points
		int n = myInput.nextInt(); // dimension of point
		double[][] points = new double[m][n];
		for (int i=0;i<m;i++) 
			for (int j=0;j<n;j++) 
				points[i][j]=myInput.nextDouble();
		myInput.close();
		kdTree kd = new kdTree(n);	
		kd.makeTree(points,0,m,0);
		points = null; // no more needed
		myInput = new Scanner(System.in);
		System.out.println("Enter the coordinates of the test point separated bt whitespace: ");
		double[] test = new double[n];
		for(int i=0;i<n;i++) 
			test[i] = myInput.nextDouble();
		showNeighbor(test);	
		System.out.println("Enter the number of nearest neighbors to find: ");
		int k = myInput.nextInt();
		myInput.close();
		if(k==1) {
			double[] nbr = NN1(kd,test);
			System.out.println("The closest neighbor is");		
			showNeighbor(nbr);
			System.out.println(String.format("distance: %4.2f",Node.dist(nbr,test)));
		} else {
			BPQ<double[]> nbrs = NNk(kd,k,test);
			System.out.println("The closest neighbors are ");
			for (double[] nbr : nbrs) { 
				showNeighbor(nbr);
				System.out.println(String.format("distance: %4.2f",Node.dist(nbr,test)));
			}
		}
	}
}
